#!/usr/bin/env python
from __future__ import print_function
import os
import sys
import re
import json
import shutil
import tempfile
import logging
from subprocess import Popen, PIPE
logging.basicConfig(level=logging.WARNING, format='%(asctime)s,%(msecs)d %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s', datefmt='%Y-%m-%d:%H:%M:%S',)
import string
import random


NXG_TOOLS_PATH = os.environ.get('NXG_TOOLS_PATH')
if NXG_TOOLS_PATH and os.path.isdir(NXG_TOOLS_PATH):
    sys.path.append(NXG_TOOLS_PATH)
elif os.path.isdir(os.path.join(os.path.dirname(__file__), '..', 'method', 'nxg-tools')):
    sys.path.append(os.path.join(os.path.dirname(__file__), '..', 'method', 'nxg-tools'))
# from nxg-tools package
from nxg_common.nxg_common import save_file_from_URI

# if the path to the TC1 executable is in an environment variable, use it
# otherwise, pull it from the path_config.py
TC1_EXECUTABLE_PATH = os.environ.get('TC1_EXECUTABLE_PATH')
if not TC1_EXECUTABLE_PATH:
    from path_config import TC1_EXECUTABLE_PATH

# if we get here and it is still not set, throw an error
if not TC1_EXECUTABLE_PATH:
    print('Path to the TC1 execuatble must be set.  See the README for details on rerunning the configure script')
    sys.exit(1)

def generate_random_str(length):
    return ''.join(random.sample(string.digits+string.ascii_letters, length))


def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)

def save_json(result, output_path):
    output_dir = os.path.dirname(output_path)
    if output_dir:
        os.makedirs(output_dir, exist_ok=True)
    with open(output_path, 'w') as w_file:
        json.dump(result, w_file, indent=2)
    return os.path.abspath(output_path)

def split_peptidea_peptideb_peptides(input_sequence_text):
    # chagne it to upper case as well
    # remove duplicates
    peptide_pair_dictkeys = dict.fromkeys(peptides for peptides in input_sequence_text.strip().upper().split('\n'))
    peptides_list = [peptides.split() for peptides in peptide_pair_dictkeys]
    return zip(*peptides_list)


def split_sequences(json_filename, parameters_output_dir=None, split_inputs_dir=None, assume_valid=False):
    logging.debug(TC1_EXECUTABLE_PATH)
    with open(json_filename, 'r') as r_file:
        input_data = json.load(r_file)

    # recreate the directory
    if os.path.exists(parameters_output_dir):
        shutil.rmtree(parameters_output_dir)

    # if not given, create a folder for it
    if not parameters_output_dir:
        parameters_output_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), os.pardir, 'job_%s/splitted_parameters' % generate_random_str(6)))
    # if not given, use parameters_output_dir for generated sequence files as well
    if not split_inputs_dir:
        split_inputs_dir = parameters_output_dir
    # create dir if not exist:
    os.makedirs(parameters_output_dir, exist_ok=True)
    os.makedirs(split_inputs_dir, exist_ok=True)

    # get splitted parameters, sequence_peptide_index table, and has_consensus flag
    #output_data, sequence_peptide_index, has_consensus = split_parameters(input_data, split_inputs_dir)

    parameters_output_dir = os.path.abspath(parameters_output_dir) 
    result_output_dir = os.path.abspath(os.path.join(parameters_output_dir, os.pardir, 'results'))
    aggregate_dir = os.path.abspath(os.path.join(parameters_output_dir, os.pardir, 'aggregate'))
    base_dir = os.path.abspath(os.path.join(parameters_output_dir, os.pardir))
    job_descriptions_path = os.path.abspath(os.path.join(base_dir, 'job_descriptions.json'))
    sequence_peptide_index_path = os.path.abspath(os.path.join(result_output_dir, 'peptides_index.json'))
    mhci_predict_executable_path = TC1_EXECUTABLE_PATH
    pvc_executable_path = os.path.abspath(os.path.join(os.path.dirname(__file__), 'run_pvc.py'))

    # create dir for peptidea and peptideb jobs
    peptidea_path = os.path.abspath(os.path.join(parameters_output_dir, 'peptidea'))
    peptideb_path = os.path.abspath(os.path.join(parameters_output_dir, 'peptideb'))
    os.makedirs(peptidea_path, exist_ok=True)
    os.makedirs(peptideb_path, exist_ok=True)

    output_data = []
    # TODO: how to deal with the  temp files, when should we remove them?
    to_delete = []
    peptide_file_path_length_pairs = []
    input_sequence_text = input_data.pop('input_sequence_text','')
    peptide_list = input_data.pop('peptide_list', '')
    input_sequence_fasta_uri = input_data.pop('input_sequence_fasta_uri', '')
    input_sequence_text_uri = input_sequence_fasta_uri.replace('sequence_list_fasta', 'download_sequences')
    if input_sequence_text_uri:
        input_sequence_text_file_path = save_file_from_URI(input_sequence_text_uri, target_dir=split_inputs_dir)
        to_delete.append(input_sequence_text_file_path)
        with open(input_sequence_text_file_path, 'r') as r_file:
            input_sequence_text = r_file.read()
        peptidea_peptide_list, peptideb_peptide_list = split_peptidea_peptideb_peptides(input_sequence_text)
    elif input_sequence_text :
        with tempfile.NamedTemporaryFile(mode='w', dir=split_inputs_dir, delete=False) as tmp_peptides_file:
            input_sequence_text_file_path = tmp_peptides_file.name
            to_delete.append(input_sequence_text_file_path)
            tmp_peptides_file.write(input_sequence_text)
        peptidea_peptide_list, peptideb_peptide_list = split_peptidea_peptideb_peptides(input_sequence_text)
    elif 'input_sequence_text_file_path' in input_data:
        input_sequence_text_file_path = input_data.pop('input_sequence_text_file_path')
        with open(input_sequence_text_file_path, 'r') as r_file:
            input_sequence_text = r_file.read()
        peptidea_peptide_list, peptideb_peptide_list = split_peptidea_peptideb_peptides(input_sequence_text)

    peptidea_peptides_path = os.path.abspath(save_peptide_list(peptidea_peptide_list, os.path.join(split_inputs_dir,'peptidea_peptides.txt')))
    peptideb_peptides_path = os.path.abspath(save_peptide_list(peptideb_peptide_list, os.path.join(split_inputs_dir,'peptideb_peptides.txt')))

    input_data['peptide_file_path'] = peptidea_peptides_path
    peptidea_input_path = save_json(input_data, os.path.join(peptidea_path, 'input.json'))
    input_data['peptide_file_path'] = peptideb_peptides_path
    peptideb_input_path = save_json(input_data, os.path.join(peptideb_path, 'input.json'))

    cmd = [TC1_EXECUTABLE_PATH, '-j', peptidea_input_path, '--split', '--split-dir=%s' % os.path.join( peptidea_path,'parameter_units'), '--keep-empty-row']
    logging.debug(' '.join(cmd))
    process = Popen(cmd, stdout=PIPE)
    cmd = [TC1_EXECUTABLE_PATH, '-j', peptideb_input_path, '--split', '--split-dir=%s' % os.path.join( peptideb_path,'parameter_units'), '--keep-empty-row']
    logging.debug(' '.join(cmd))
    process = Popen(cmd, stdout=PIPE)
    stdoutdata, stderrdata_ignored = process.communicate()
    stdoutdata = stdoutdata.decode()
    logging.debug('Raw output:\n{}'.format(stdoutdata))

    peptidea_job_descriptions_json_file = os.path.join(peptidea_path, 'job_descriptions.json')
    peptideb_job_descriptions_json_file = os.path.join(peptideb_path, 'job_descriptions.json')
    final_job_descriptions_json_file = os.path.abspath(os.path.join(parameters_output_dir, '..', 'job_descriptions.json'))
    final_result_path = os.path.join(parameters_output_dir, 'aggregated_result.json')

    peptidea_job_descriptions = read_json_file(peptidea_job_descriptions_json_file)
    peptideb_job_descriptions = read_json_file(peptideb_job_descriptions_json_file)
    final_job_descriptions = merge_jobs_and_add_aggregate(peptidea_job_descriptions,peptideb_job_descriptions, pvc_executable_path, final_result_path)
    save_json(final_job_descriptions, final_job_descriptions_json_file)

    print(f"job_descriptions_json_file: {final_job_descriptions_json_file}")
    return final_job_descriptions_json_file

def merge_jobs_and_add_aggregate(peptidea_job_descriptions, peptideb_job_descriptions ,pvc_executable_path, final_result_path):
    final_job_descriptions = []
    final_aggregate_depends_on_job_ids = []

    peptidea_output_path, peptidea_final_job_id = add_job_descriptions(peptidea_job_descriptions, final_job_descriptions)
    final_aggregate_depends_on_job_ids.append(peptidea_final_job_id)

    peptideb_output_path, peptideb_final_job_id = add_job_descriptions(peptideb_job_descriptions, final_job_descriptions)
    final_aggregate_depends_on_job_ids.append(peptideb_final_job_id)

    final_aggregate_cmd = (pvc_executable_path, "--aggregate", f"--peptidea-result-path={peptidea_output_path}", f"--peptideb-result-path={peptideb_output_path}", f"--aggregate-output-path={final_result_path}")
    final_job_id = max([job['job_id'] for job in final_job_descriptions]) + 1

    final_job_descriptions.append({
        "shell_cmd": " ".join(final_aggregate_cmd),
        "job_id": final_job_id,
        "job_type": "aggregate",
        "depends_on_job_ids": final_aggregate_depends_on_job_ids,
        "expected_outputs": [final_result_path,]
    })
    return final_job_descriptions

def add_job_descriptions(sub_job_descriptions, final_job_descriptions):
    initial_job_id = 1
    if final_job_descriptions:
        initial_job_id = max([job['job_id'] for job in final_job_descriptions]) + 1
    new_job_id = initial_job_id
    job_id_map = {}
    for job in sub_job_descriptions:
        job_id_map[job['job_id']] = new_job_id
        job['job_id'] = new_job_id
        job['depends_on_job_ids'] = [job_id_map[old_id] for old_id in job['depends_on_job_ids']]
        final_job_descriptions.append(job)
        if job["job_type"] == "aggregate":
            output_path = job['expected_outputs'][0]
            aggregate_job_id = job['job_id']
        new_job_id += 1
    return output_path, aggregate_job_id

def save_peptide_list(peptide_list, file_path):
    with open(file_path, mode='w') as peptides_file:
        peptides_file.write('\n'.join(peptide_list))
    return file_path

def read_json_file(file_path):
    with open(file_path, 'r') as r_file:
        return json.load(r_file)

if __name__ == '__main__':
    pass


